@@ -24,15 +24,37 @@ import (
 	"sync"
 	"time"
 
-	"github.com/go-kit/kit/log"
-	"github.com/go-kit/kit/log/level"
+	"github.com/go-kit/log"
+	"github.com/go-kit/log/level"
 	"github.com/hashicorp/memberlist"
 	"github.com/oklog/ulid"
 	"github.com/pkg/errors"
-
 	"github.com/prometheus/client_golang/prometheus"
 )
 
+// ClusterPeer represents a single Peer in a gossip cluster.
+type ClusterPeer interface {
+	// Name returns the unique identifier of this peer in the cluster.
+	Name() string
+	// Status returns a status string representing the peer state.
+	Status() string
+	// Peers returns the peer nodes in the cluster.
+	Peers() []ClusterMember
+}
+
+// ClusterMember interface that represents node peers in a cluster
+type ClusterMember interface {
+	// Name returns the name of the node
+	Name() string
+	// Address returns the IP address of the node
+	Address() string
+}
+
+// ClusterChannel supports state broadcasting across peers.
+type ClusterChannel interface {
+	Broadcast([]byte)
+}
+
 // Peer is a single peer in a gossip cluster.
 type Peer struct {
 	mlist    *memberlist.Memberlist
@@ -97,13 +119,13 @@ func (s PeerStatus) String() string {
 const (
 	DefaultPushPullInterval  = 60 * time.Second
 	DefaultGossipInterval    = 200 * time.Millisecond
-	DefaultTcpTimeout        = 10 * time.Second
+	DefaultTCPTimeout        = 10 * time.Second
 	DefaultProbeTimeout      = 500 * time.Millisecond
 	DefaultProbeInterval     = 1 * time.Second
 	DefaultReconnectInterval = 10 * time.Second
 	DefaultReconnectTimeout  = 6 * time.Hour
 	DefaultRefreshInterval   = 15 * time.Second
-	maxGossipPacketSize      = 1400
+	MaxGossipPacketSize      = 1400
 )
 
 func Create(
@@ -118,6 +140,9 @@ func Create(
 	tcpTimeout time.Duration,
 	probeTimeout time.Duration,
 	probeInterval time.Duration,
+	tlsTransportConfig *TLSTransportConfig,
+	allowInsecureAdvertise bool,
+	label string,
 ) (*Peer, error) {
 	bindHost, bindPortStr, err := net.SplitHostPort(bindAddr)
 	if err != nil {
@@ -149,7 +174,7 @@ func Create(
 	level.Debug(l).Log("msg", "resolved peers to following addresses", "peers", strings.Join(resolvedPeers, ","))
 
 	// Initial validation of user-specified advertise address.
-	addr, err := calculateAdvertiseAddress(bindHost, advertiseHost)
+	addr, err := calculateAdvertiseAddress(bindHost, advertiseHost, allowInsecureAdvertise)
 	if err != nil {
 		level.Warn(l).Log("err", "couldn't deduce an advertise address: "+err.Error())
 	} else if hasNonlocal(resolvedPeers) && isUnroutable(addr.String()) {
@@ -202,7 +227,8 @@ func Create(
 	cfg.ProbeInterval = probeInterval
 	cfg.LogOutput = &logWriter{l: l}
 	cfg.GossipNodes = retransmit
-	cfg.UDPBufferSize = maxGossipPacketSize
+	cfg.UDPBufferSize = MaxGossipPacketSize
+	cfg.Label = label
 
 	if advertiseHost != "" {
 		cfg.AdvertiseAddr = advertiseHost
@@ -212,6 +238,14 @@ func Create(
 		p.setInitialFailed(resolvedPeers, bindAddr)
 	}
 
+	if tlsTransportConfig != nil {
+		level.Info(l).Log("msg", "using TLS for gossip")
+		cfg.Transport, err = NewTLSTransport(context.Background(), l, reg, cfg.BindAddr, cfg.BindPort, tlsTransportConfig)
+		if err != nil {
+			return nil, errors.Wrap(err, "tls transport")
+		}
+	}
+
 	ml, err := memberlist.Create(cfg)
 	if err != nil {
 		return nil, errors.Wrap(err, "create memberlist")
@@ -222,7 +256,8 @@ func Create(
 
 func (p *Peer) Join(
 	reconnectInterval time.Duration,
-	reconnectTimeout time.Duration) error {
+	reconnectTimeout time.Duration,
+) error {
 	n, err := p.mlist.Join(p.resolvedPeers)
 	if err != nil {
 		level.Warn(p.logger).Log("msg", "failed to join cluster", "err", err)
@@ -512,15 +547,18 @@ func (p *Peer) peerUpdate(n *memberlist.Node) {
 
 // AddState adds a new state that will be gossiped. It returns a channel to which
 // broadcast messages for the state can be sent.
-func (p *Peer) AddState(key string, s State, reg prometheus.Registerer) *Channel {
+func (p *Peer) AddState(key string, s State, reg prometheus.Registerer) ClusterChannel {
+	p.mtx.Lock()
 	p.states[key] = s
+	p.mtx.Unlock()
+
 	send := func(b []byte) {
 		p.delegate.bcast.QueueBroadcast(simpleBroadcast(b))
 	}
 	peers := func() []*memberlist.Node {
-		nodes := p.Peers()
+		nodes := p.mlist.Members()
 		for i, n := range nodes {
-			if n.Name == p.Self().Name {
+			if n.String() == p.Self().Name {
 				nodes = append(nodes[:i], nodes[i+1:]...)
 				break
 			}
@@ -561,8 +599,13 @@ func (p *Peer) Ready() bool {
 }
 
 // Wait until Settle() has finished.
-func (p *Peer) WaitReady() {
-	<-p.readyc
+func (p *Peer) WaitReady(ctx context.Context) error {
+	select {
+	case <-ctx.Done():
+		return ctx.Err()
+	case <-p.readyc:
+		return nil
+	}
 }
 
 // Return a status string representing the peer state.
@@ -591,14 +634,31 @@ func (p *Peer) Self() *memberlist.Node {
 	return p.mlist.LocalNode()
 }
 
+// Member represents a member in the cluster.
+type Member struct {
+	node *memberlist.Node
+}
+
+// Name implements cluster.ClusterMember
+func (m Member) Name() string { return m.node.Name }
+
+// Address implements cluster.ClusterMember
+func (m Member) Address() string { return m.node.Address() }
+
 // Peers returns the peers in the cluster.
-func (p *Peer) Peers() []*memberlist.Node {
-	return p.mlist.Members()
+func (p *Peer) Peers() []ClusterMember {
+	peers := make([]ClusterMember, 0, len(p.mlist.Members()))
+	for _, member := range p.mlist.Members() {
+		peers = append(peers, Member{
+			node: member,
+		})
+	}
+	return peers
 }
 
 // Position returns the position of the peer in the cluster.
 func (p *Peer) Position() int {
-	all := p.Peers()
+	all := p.mlist.Members()
 	sort.Slice(all, func(i, j int) bool {
 		return all[i].Name < all[j].Name
 	})
@@ -727,7 +787,7 @@ func resolvePeers(ctx context.Context, peers []string, myAddress string, res *ne
 	return resolvedPeers, nil
 }
 
-func removeMyAddr(ips []net.IPAddr, targetPort string, myAddr string) []net.IPAddr {
+func removeMyAddr(ips []net.IPAddr, targetPort, myAddr string) []net.IPAddr {
 	var result []net.IPAddr
 
 	for _, ip := range ips {
