@@ -15,20 +15,43 @@ package silence
 
 import (
 	"bytes"
-	"io/ioutil"
+	"fmt"
 	"os"
+	"runtime"
 	"sort"
-	"strings"
+	"sync"
 	"testing"
 	"time"
 
+	"github.com/benbjohnson/clock"
+	"github.com/go-kit/log"
 	"github.com/matttproud/golang_protobuf_extensions/pbutil"
-	pb "github.com/prometheus/alertmanager/silence/silencepb"
-	"github.com/prometheus/alertmanager/types"
+	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/client_golang/prometheus/testutil"
 	"github.com/prometheus/common/model"
 	"github.com/stretchr/testify/require"
+	"go.uber.org/atomic"
+
+	pb "github.com/prometheus/alertmanager/silence/silencepb"
+	"github.com/prometheus/alertmanager/types"
 )
 
+func checkErr(t *testing.T, expected string, got error) {
+	t.Helper()
+
+	if expected == "" {
+		require.NoError(t, got)
+		return
+	}
+
+	if got == nil {
+		t.Errorf("expected error containing %q but got none", expected)
+		return
+	}
+
+	require.Contains(t, got.Error(), expected)
+}
+
 func TestOptionsValidate(t *testing.T) {
 	cases := []struct {
 		options *Options
@@ -54,20 +77,7 @@ func TestOptionsValidate(t *testing.T) {
 	}
 
 	for _, c := range cases {
-		err := c.options.validate()
-		if err == nil {
-			if c.err != "" {
-				t.Errorf("expected error containing %q but got none", c.err)
-			}
-			continue
-		}
-		if err != nil && c.err == "" {
-			t.Errorf("unexpected error %q", err)
-			continue
-		}
-		if !strings.Contains(err.Error(), c.err) {
-			t.Errorf("expected error to contain %q but got %q", c.err, err)
-		}
+		checkErr(t, c.err, c.options.validate())
 	}
 }
 
@@ -75,8 +85,8 @@ func TestSilencesGC(t *testing.T) {
 	s, err := New(Options{})
 	require.NoError(t, err)
 
-	now := utcNow()
-	s.now = func() time.Time { return now }
+	s.clock = clock.NewMock()
+	now := s.nowUTC()
 
 	newSilence := func(exp time.Time) *pb.MeshSilence {
 		return &pb.MeshSilence{ExpiresAt: exp}
@@ -98,7 +108,7 @@ func TestSilencesGC(t *testing.T) {
 
 func TestSilencesSnapshot(t *testing.T) {
 	// Check whether storing and loading the snapshot is symmetric.
-	now := utcNow()
+	now := clock.NewMock().Now().UTC()
 
 	cases := []struct {
 		entries []*pb.MeshSilence
@@ -118,6 +128,19 @@ func TestSilencesSnapshot(t *testing.T) {
 					},
 					ExpiresAt: now,
 				},
+				{
+					Silence: &pb.Silence{
+						Id: "3dfb2528-59ce-41eb-b465-f875a4e744a4",
+						Matchers: []*pb.Matcher{
+							{Name: "label1", Pattern: "val1", Type: pb.Matcher_NOT_EQUAL},
+							{Name: "label2", Pattern: "val.+", Type: pb.Matcher_NOT_REGEXP},
+						},
+						StartsAt:  now,
+						EndsAt:    now,
+						UpdatedAt: now,
+					},
+					ExpiresAt: now,
+				},
 				{
 					Silence: &pb.Silence{
 						Id: "4b1e760d-182c-4980-b873-c1a6827c9817",
@@ -135,7 +158,7 @@ func TestSilencesSnapshot(t *testing.T) {
 	}
 
 	for _, c := range cases {
-		f, err := ioutil.TempFile("", "snapshot")
+		f, err := os.CreateTemp("", "snapshot")
 		require.NoError(t, err, "creating temp file failed")
 
 		s1 := &Silences{st: state{}, metrics: newMetrics(nil, nil)}
@@ -161,14 +184,84 @@ func TestSilencesSnapshot(t *testing.T) {
 	}
 }
 
+// This tests a regression introduced by https://github.com/prometheus/alertmanager/pull/2689.
+func TestSilences_Maintenance_DefaultMaintenanceFuncDoesntCrash(t *testing.T) {
+	f, err := os.CreateTemp("", "snapshot")
+	require.NoError(t, err, "creating temp file failed")
+	clock := clock.NewMock()
+	s := &Silences{st: state{}, logger: log.NewNopLogger(), clock: clock, metrics: newMetrics(nil, nil)}
+	stopc := make(chan struct{})
+
+	done := make(chan struct{})
+	go func() {
+		s.Maintenance(100*time.Millisecond, f.Name(), stopc, nil)
+		close(done)
+	}()
+	runtime.Gosched()
+
+	clock.Add(100 * time.Millisecond)
+	close(stopc)
+
+	<-done
+}
+
+func TestSilences_Maintenance_SupportsCustomCallback(t *testing.T) {
+	f, err := os.CreateTemp("", "snapshot")
+	require.NoError(t, err, "creating temp file failed")
+	clock := clock.NewMock()
+	reg := prometheus.NewRegistry()
+	s := &Silences{st: state{}, logger: log.NewNopLogger(), clock: clock}
+	s.metrics = newMetrics(reg, s)
+	stopc := make(chan struct{})
+
+	var calls atomic.Int32
+	var wg sync.WaitGroup
+
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+		s.Maintenance(10*time.Second, f.Name(), stopc, func() (int64, error) {
+			calls.Add(1)
+			return 0, nil
+		})
+	}()
+	gosched()
+
+	// Before the first tick, no maintenance executed.
+	clock.Add(9 * time.Second)
+	require.EqualValues(t, 0, calls.Load())
+
+	// Tick once.
+	clock.Add(1 * time.Second)
+	require.EqualValues(t, 1, calls.Load())
+
+	// Stop the maintenance loop. We should get exactly one more execution of the maintenance func.
+	close(stopc)
+	wg.Wait()
+
+	require.EqualValues(t, 2, calls.Load())
+
+	// Check the maintenance metrics.
+	require.NoError(t, testutil.GatherAndCompare(reg, bytes.NewBufferString(`
+# HELP alertmanager_silences_maintenance_errors_total How many maintenances were executed for silences that failed.
+# TYPE alertmanager_silences_maintenance_errors_total counter
+alertmanager_silences_maintenance_errors_total 0
+# HELP alertmanager_silences_maintenance_total How many maintenances were executed for silences.
+# TYPE alertmanager_silences_maintenance_total counter
+alertmanager_silences_maintenance_total 2
+`), "alertmanager_silences_maintenance_total", "alertmanager_silences_maintenance_errors_total"))
+}
+
 func TestSilencesSetSilence(t *testing.T) {
 	s, err := New(Options{
 		Retention: time.Minute,
 	})
 	require.NoError(t, err)
 
-	now := utcNow()
-	nowpb := now
+	clock := clock.NewMock()
+	s.clock = clock
+
+	nowpb := s.nowUTC()
 
 	sil := &pb.Silence{
 		Id:       "some_id",
@@ -180,11 +273,11 @@ func TestSilencesSetSilence(t *testing.T) {
 	want := state{
 		"some_id": &pb.MeshSilence{
 			Silence:   sil,
-			ExpiresAt: now.Add(time.Minute),
+			ExpiresAt: nowpb.Add(time.Minute),
 		},
 	}
 
-	done := make(chan bool)
+	done := make(chan struct{})
 	s.broadcast = func(b []byte) {
 		var e pb.MeshSilence
 		r := bytes.NewReader(b)
@@ -195,18 +288,16 @@ func TestSilencesSetSilence(t *testing.T) {
 		close(done)
 	}
 
-	// setSilence() is always called with s.mtx locked()
-	go func() {
+	// setSilence() is always called with s.mtx locked() in the application code
+	func() {
 		s.mtx.Lock()
-		require.NoError(t, s.setSilence(sil))
-		s.mtx.Unlock()
+		defer s.mtx.Unlock()
+		require.NoError(t, s.setSilence(sil, nowpb))
 	}()
 
-	// GossipBroadcast is called in a goroutine.
-	select {
-	case <-done:
-	case <-time.After(1 * time.Second):
-		t.Fatal("GossipBroadcast was not called")
+	// Ensure broadcast was called.
+	if _, isOpen := <-done; isOpen {
+		t.Fatal("broadcast was not called")
 	}
 
 	require.Equal(t, want, s.st, "Unexpected silence state")
@@ -218,15 +309,15 @@ func TestSilenceSet(t *testing.T) {
 	})
 	require.NoError(t, err)
 
-	now := utcNow()
-	now1 := now
-	s.now = func() time.Time { return now }
+	clock := clock.NewMock()
+	s.clock = clock
+	start1 := s.nowUTC()
 
 	// Insert silence with fixed start time.
 	sil1 := &pb.Silence{
 		Matchers: []*pb.Matcher{{Name: "a", Pattern: "b"}},
-		StartsAt: now.Add(2 * time.Minute),
-		EndsAt:   now.Add(5 * time.Minute),
+		StartsAt: start1.Add(2 * time.Minute),
+		EndsAt:   start1.Add(5 * time.Minute),
 	}
 	id1, err := s.Set(sil1)
 	require.NoError(t, err)
@@ -237,22 +328,22 @@ func TestSilenceSet(t *testing.T) {
 			Silence: &pb.Silence{
 				Id:        id1,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now1.Add(2 * time.Minute),
-				EndsAt:    now1.Add(5 * time.Minute),
-				UpdatedAt: now1,
+				StartsAt:  start1.Add(2 * time.Minute),
+				EndsAt:    start1.Add(5 * time.Minute),
+				UpdatedAt: start1,
 			},
-			ExpiresAt: now1.Add(5*time.Minute + s.retention),
+			ExpiresAt: start1.Add(5*time.Minute + s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
 
 	// Insert silence with unset start time. Must be set to now.
-	now = now.Add(time.Minute)
-	now2 := now
+	clock.Add(time.Minute)
+	start2 := s.nowUTC()
 
 	sil2 := &pb.Silence{
 		Matchers: []*pb.Matcher{{Name: "a", Pattern: "b"}},
-		EndsAt:   now.Add(1 * time.Minute),
+		EndsAt:   start2.Add(1 * time.Minute),
 	}
 	id2, err := s.Set(sil2)
 	require.NoError(t, err)
@@ -264,21 +355,21 @@ func TestSilenceSet(t *testing.T) {
 			Silence: &pb.Silence{
 				Id:        id2,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now2,
-				EndsAt:    now2.Add(1 * time.Minute),
-				UpdatedAt: now2,
+				StartsAt:  start2,
+				EndsAt:    start2.Add(1 * time.Minute),
+				UpdatedAt: start2,
 			},
-			ExpiresAt: now2.Add(1*time.Minute + s.retention),
+			ExpiresAt: start2.Add(1*time.Minute + s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
 
 	// Overwrite silence 2 with new end time.
-	now = now.Add(time.Minute)
-	now3 := now
+	clock.Add(time.Minute)
+	start3 := s.nowUTC()
 
 	sil3 := cloneSilence(sil2)
-	sil3.EndsAt = now.Add(100 * time.Minute)
+	sil3.EndsAt = start3.Add(100 * time.Minute)
 
 	id3, err := s.Set(sil3)
 	require.NoError(t, err)
@@ -290,24 +381,25 @@ func TestSilenceSet(t *testing.T) {
 			Silence: &pb.Silence{
 				Id:        id2,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now2,
-				EndsAt:    now3.Add(100 * time.Minute),
-				UpdatedAt: now3,
+				StartsAt:  start2,
+				EndsAt:    start3.Add(100 * time.Minute),
+				UpdatedAt: start3,
 			},
-			ExpiresAt: now3.Add(100*time.Minute + s.retention),
+			ExpiresAt: start3.Add(100*time.Minute + s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
 
-	// Update silence 2 with new matcher expires it and creates a new one.
-	now = now.Add(time.Minute)
-	now4 := now
+	// Update this silence again with new matcher. This expires it and creates a new one.
+	clock.Add(time.Minute)
+	start4 := s.nowUTC()
 
 	sil4 := cloneSilence(sil3)
 	sil4.Matchers = []*pb.Matcher{{Name: "a", Pattern: "c"}}
 
 	id4, err := s.Set(sil4)
 	require.NoError(t, err)
+	// This new silence gets a new id.
 	require.NotEqual(t, id2, id4)
 
 	want = state{
@@ -316,32 +408,32 @@ func TestSilenceSet(t *testing.T) {
 			Silence: &pb.Silence{
 				Id:        id2,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now2,
-				EndsAt:    now4,
-				UpdatedAt: now4,
+				StartsAt:  start2,
+				EndsAt:    start4, // Expired
+				UpdatedAt: start4,
 			},
-			ExpiresAt: now4.Add(s.retention),
+			ExpiresAt: start4.Add(s.retention),
 		},
 		id4: &pb.MeshSilence{
 			Silence: &pb.Silence{
 				Id:        id4,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "c"}},
-				StartsAt:  now4,
-				EndsAt:    now3.Add(100 * time.Minute),
-				UpdatedAt: now4,
+				StartsAt:  start4,
+				EndsAt:    start3.Add(100 * time.Minute),
+				UpdatedAt: start4,
 			},
-			ExpiresAt: now3.Add(100*time.Minute + s.retention),
+			ExpiresAt: start3.Add(100*time.Minute + s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
 
-	// Re-create expired silence.
-	now = now.Add(time.Minute)
-	now5 := now
+	// Re-create the silence that just expired.
+	clock.Add(time.Minute)
+	start5 := s.nowUTC()
 
 	sil5 := cloneSilence(sil3)
-	sil5.StartsAt = now
-	sil5.EndsAt = now.Add(5 * time.Minute)
+	sil5.StartsAt = start1
+	sil5.EndsAt = start1.Add(5 * time.Minute)
 
 	id5, err := s.Set(sil5)
 	require.NoError(t, err)
@@ -355,11 +447,62 @@ func TestSilenceSet(t *testing.T) {
 			Silence: &pb.Silence{
 				Id:        id5,
 				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
-				StartsAt:  now5,
-				EndsAt:    now5.Add(5 * time.Minute),
-				UpdatedAt: now5,
+				StartsAt:  start5, // New silences have their start time set to "now" when created.
+				EndsAt:    start1.Add(5 * time.Minute),
+				UpdatedAt: start5,
 			},
-			ExpiresAt: now5.Add(5*time.Minute + s.retention),
+			ExpiresAt: start1.Add(5*time.Minute + s.retention),
+		},
+	}
+	require.Equal(t, want, s.st, "unexpected state after silence creation")
+}
+
+func TestSetActiveSilence(t *testing.T) {
+	s, err := New(Options{
+		Retention: time.Hour,
+	})
+	require.NoError(t, err)
+
+	clock := clock.NewMock()
+	s.clock = clock
+	now := clock.Now()
+
+	startsAt := now.Add(-1 * time.Minute)
+	endsAt := now.Add(5 * time.Minute)
+	// Insert silence with fixed start time.
+	sil1 := &pb.Silence{
+		Matchers: []*pb.Matcher{{Name: "a", Pattern: "b"}},
+		StartsAt: startsAt,
+		EndsAt:   endsAt,
+	}
+	id1, _ := s.Set(sil1)
+
+	// Update silence with 2 extra nanoseconds so the "seconds" part should not change
+
+	newStartsAt := now.Add(2 * time.Nanosecond)
+	newEndsAt := endsAt.Add(2 * time.Minute)
+
+	sil2 := cloneSilence(sil1)
+	sil2.Id = id1
+	sil2.StartsAt = newStartsAt
+	sil2.EndsAt = newEndsAt
+
+	clock.Add(time.Minute)
+	now = s.nowUTC()
+	id2, err := s.Set(sil2)
+	require.NoError(t, err)
+	require.Equal(t, id1, id2)
+
+	want := state{
+		id2: &pb.MeshSilence{
+			Silence: &pb.Silence{
+				Id:        id1,
+				Matchers:  []*pb.Matcher{{Name: "a", Pattern: "b"}},
+				StartsAt:  newStartsAt,
+				EndsAt:    newEndsAt,
+				UpdatedAt: now,
+			},
+			ExpiresAt: newEndsAt.Add(s.retention),
 		},
 	}
 	require.Equal(t, want, s.st, "unexpected state after silence creation")
@@ -369,8 +512,8 @@ func TestSilencesSetFail(t *testing.T) {
 	s, err := New(Options{})
 	require.NoError(t, err)
 
-	now := utcNow()
-	s.now = func() time.Time { return now }
+	clock := clock.NewMock()
+	s.clock = clock
 
 	cases := []struct {
 		s   *pb.Silence
@@ -386,24 +529,12 @@ func TestSilencesSetFail(t *testing.T) {
 	}
 	for _, c := range cases {
 		_, err := s.Set(c.s)
-		if err == nil {
-			if c.err != "" {
-				t.Errorf("expected error containing %q but got none", c.err)
-			}
-			continue
-		}
-		if err != nil && c.err == "" {
-			t.Errorf("unexpected error %q", err)
-			continue
-		}
-		if !strings.Contains(err.Error(), c.err) {
-			t.Errorf("expected error to contain %q but got %q", c.err, err)
-		}
+		checkErr(t, c.err, err)
 	}
 }
 
 func TestQState(t *testing.T) {
-	now := utcNow()
+	now := time.Now().UTC()
 
 	cases := []struct {
 		sil    *pb.Silence
@@ -470,6 +601,14 @@ func TestQMatches(t *testing.T) {
 			},
 			drop: true,
 		},
+		{
+			sil: &pb.Silence{
+				Matchers: []*pb.Matcher{
+					{Name: "job", Pattern: "test", Type: pb.Matcher_NOT_EQUAL},
+				},
+			},
+			drop: false,
+		},
 		{
 			sil: &pb.Silence{
 				Matchers: []*pb.Matcher{
@@ -479,6 +618,15 @@ func TestQMatches(t *testing.T) {
 			},
 			drop: false,
 		},
+		{
+			sil: &pb.Silence{
+				Matchers: []*pb.Matcher{
+					{Name: "job", Pattern: "test", Type: pb.Matcher_EQUAL},
+					{Name: "method", Pattern: "POST", Type: pb.Matcher_NOT_EQUAL},
+				},
+			},
+			drop: true,
+		},
 		{
 			sil: &pb.Silence{
 				Matchers: []*pb.Matcher{
@@ -487,6 +635,14 @@ func TestQMatches(t *testing.T) {
 			},
 			drop: true,
 		},
+		{
+			sil: &pb.Silence{
+				Matchers: []*pb.Matcher{
+					{Name: "path", Pattern: "/user/.+", Type: pb.Matcher_NOT_REGEXP},
+				},
+			},
+			drop: false,
+		},
 		{
 			sil: &pb.Silence{
 				Matchers: []*pb.Matcher{
@@ -572,7 +728,7 @@ func TestSilencesQuery(t *testing.T) {
 
 	for _, c := range cases {
 		// Run default query of retrieving all silences.
-		res, err := s.query(c.q, time.Time{})
+		res, _, err := s.query(c.q, time.Time{})
 		require.NoError(t, err, "unexpected error on querying")
 
 		// Currently there are no sorting guarantees in the querying API.
@@ -589,7 +745,7 @@ func (s silencesByID) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
 func (s silencesByID) Less(i, j int) bool { return s[i].Id < s[j].Id }
 
 func TestSilenceCanUpdate(t *testing.T) {
-	now := utcNow()
+	now := time.Now().UTC()
 
 	cases := []struct {
 		a, b *pb.Silence
@@ -715,11 +871,12 @@ func TestSilenceCanUpdate(t *testing.T) {
 }
 
 func TestSilenceExpire(t *testing.T) {
-	s, err := New(Options{})
+	s, err := New(Options{Retention: time.Hour})
 	require.NoError(t, err)
 
-	now := time.Now()
-	s.now = func() time.Time { return now }
+	clock := clock.NewMock()
+	s.clock = clock
+	now := s.nowUTC()
 
 	m := &pb.Matcher{Type: pb.Matcher_EQUAL, Name: "a", Pattern: "b"}
 
@@ -751,12 +908,14 @@ func TestSilenceExpire(t *testing.T) {
 	require.NoError(t, err)
 	require.Equal(t, 1, count)
 
-	require.NoError(t, s.expire("pending"))
-	require.NoError(t, s.expire("active"))
+	count, err = s.CountState(types.SilenceStateExpired)
+	require.NoError(t, err)
+	require.Equal(t, 1, count)
 
-	err = s.expire("expired")
-	require.Error(t, err)
-	require.Contains(t, err.Error(), "already expired")
+	require.NoError(t, s.Expire("pending"))
+	require.NoError(t, s.Expire("active"))
+
+	require.NoError(t, s.Expire("expired"))
 
 	sil, err := s.QueryOne(QIDs("pending"))
 	require.NoError(t, err)
@@ -768,10 +927,17 @@ func TestSilenceExpire(t *testing.T) {
 		UpdatedAt: now,
 	}, sil)
 
+	// Let time pass...
+	clock.Add(time.Second)
+
 	count, err = s.CountState(types.SilenceStatePending)
 	require.NoError(t, err)
 	require.Equal(t, 0, count)
 
+	count, err = s.CountState(types.SilenceStateExpired)
+	require.NoError(t, err)
+	require.Equal(t, 3, count)
+
 	// Expiring a pending Silence should make the API return the
 	// SilenceStateExpired Silence state.
 	silenceState := types.CalcSilenceState(sil.StartsAt, sil.EndsAt)
@@ -798,6 +964,156 @@ func TestSilenceExpire(t *testing.T) {
 	}, sil)
 }
 
+// TestSilenceExpireWithZeroRetention covers the problem that, with zero
+// retention time, a silence explicitly set to expired will also immediately
+// expire from the silence storage.
+func TestSilenceExpireWithZeroRetention(t *testing.T) {
+	s, err := New(Options{Retention: 0})
+	require.NoError(t, err)
+
+	clock := clock.NewMock()
+	s.clock = clock
+	now := s.nowUTC()
+
+	m := &pb.Matcher{Type: pb.Matcher_EQUAL, Name: "a", Pattern: "b"}
+
+	s.st = state{
+		"pending": &pb.MeshSilence{Silence: &pb.Silence{
+			Id:        "pending",
+			Matchers:  []*pb.Matcher{m},
+			StartsAt:  now.Add(time.Minute),
+			EndsAt:    now.Add(time.Hour),
+			UpdatedAt: now.Add(-time.Hour),
+		}},
+		"active": &pb.MeshSilence{Silence: &pb.Silence{
+			Id:        "active",
+			Matchers:  []*pb.Matcher{m},
+			StartsAt:  now.Add(-time.Minute),
+			EndsAt:    now.Add(time.Hour),
+			UpdatedAt: now.Add(-time.Hour),
+		}},
+		"expired": &pb.MeshSilence{Silence: &pb.Silence{
+			Id:        "expired",
+			Matchers:  []*pb.Matcher{m},
+			StartsAt:  now.Add(-time.Hour),
+			EndsAt:    now.Add(-time.Minute),
+			UpdatedAt: now.Add(-time.Hour),
+		}},
+	}
+
+	count, err := s.CountState(types.SilenceStatePending)
+	require.NoError(t, err)
+	require.Equal(t, 1, count)
+
+	count, err = s.CountState(types.SilenceStateActive)
+	require.NoError(t, err)
+	require.Equal(t, 1, count)
+
+	count, err = s.CountState(types.SilenceStateExpired)
+	require.NoError(t, err)
+	require.Equal(t, 1, count)
+
+	// Advance time. The silence state management code uses update time when
+	// merging, and the logic is "first write wins". So we must advance the clock
+	// one tick for updates to take effect.
+	clock.Add(1 * time.Millisecond)
+
+	require.NoError(t, s.Expire("pending"))
+	require.NoError(t, s.Expire("active"))
+	require.NoError(t, s.Expire("expired"))
+
+	// Advance time again. Despite what the function name says, s.Expire() does
+	// not expire a silence. It sets the silence to EndAt the current time. This
+	// means that the silence is active immediately after calling Expire.
+	clock.Add(1 * time.Millisecond)
+
+	// Verify all silences have expired.
+	count, err = s.CountState(types.SilenceStatePending)
+	require.NoError(t, err)
+	require.Equal(t, 0, count)
+
+	count, err = s.CountState(types.SilenceStateActive)
+	require.NoError(t, err)
+	require.Equal(t, 0, count)
+
+	count, err = s.CountState(types.SilenceStateExpired)
+	require.NoError(t, err)
+	require.Equal(t, 3, count)
+}
+
+func TestSilencer(t *testing.T) {
+	ss, err := New(Options{Retention: time.Hour})
+	require.NoError(t, err)
+
+	clock := clock.NewMock()
+	ss.clock = clock
+	now := ss.nowUTC()
+
+	m := types.NewMarker(prometheus.NewRegistry())
+	s := NewSilencer(ss, m, log.NewNopLogger())
+
+	require.False(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert not silenced without any silences")
+
+	_, err = ss.Set(&pb.Silence{
+		Matchers: []*pb.Matcher{{Name: "foo", Pattern: "baz"}},
+		StartsAt: now.Add(-time.Hour),
+		EndsAt:   now.Add(5 * time.Minute),
+	})
+	require.NoError(t, err)
+
+	require.False(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert not silenced by non-matching silence")
+
+	id, err := ss.Set(&pb.Silence{
+		Matchers: []*pb.Matcher{{Name: "foo", Pattern: "bar"}},
+		StartsAt: now.Add(-time.Hour),
+		EndsAt:   now.Add(5 * time.Minute),
+	})
+	require.NoError(t, err)
+	require.NotEmpty(t, id)
+
+	require.True(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert silenced by matching silence")
+
+	// One hour passes, silence expires.
+	clock.Add(time.Hour)
+	now = ss.nowUTC()
+
+	require.False(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert not silenced by expired silence")
+
+	// Update silence to start in the future.
+	_, err = ss.Set(&pb.Silence{
+		Id:       id,
+		Matchers: []*pb.Matcher{{Name: "foo", Pattern: "bar"}},
+		StartsAt: now.Add(time.Hour),
+		EndsAt:   now.Add(3 * time.Hour),
+	})
+	require.NoError(t, err)
+
+	require.False(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert not silenced by future silence")
+
+	// Two hours pass, silence becomes active.
+	clock.Add(2 * time.Hour)
+	now = ss.nowUTC()
+
+	// Exposes issue #2426.
+	require.True(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert silenced by activated silence")
+
+	_, err = ss.Set(&pb.Silence{
+		Matchers: []*pb.Matcher{{Name: "foo", Pattern: "b..", Type: pb.Matcher_REGEXP}},
+		StartsAt: now.Add(time.Hour),
+		EndsAt:   now.Add(3 * time.Hour),
+	})
+	require.NoError(t, err)
+
+	// Note that issue #2426 doesn't apply anymore because we added a new silence.
+	require.True(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert still silenced by activated silence")
+
+	// Two hours pass, first silence expires, overlapping second silence becomes active.
+	clock.Add(2 * time.Hour)
+
+	// Another variant of issue #2426 (overlapping silences).
+	require.True(t, s.Mutes(model.LabelSet{"foo": "bar"}), "expected alert silenced by activated second silence")
+}
+
 func TestValidateMatcher(t *testing.T) {
 	cases := []struct {
 		m   *pb.Matcher
@@ -810,6 +1126,27 @@ func TestValidateMatcher(t *testing.T) {
 				Type:    pb.Matcher_EQUAL,
 			},
 			err: "",
+		}, {
+			m: &pb.Matcher{
+				Name:    "a",
+				Pattern: "b",
+				Type:    pb.Matcher_NOT_EQUAL,
+			},
+			err: "",
+		}, {
+			m: &pb.Matcher{
+				Name:    "a",
+				Pattern: "b",
+				Type:    pb.Matcher_REGEXP,
+			},
+			err: "",
+		}, {
+			m: &pb.Matcher{
+				Name:    "a",
+				Pattern: "b",
+				Type:    pb.Matcher_NOT_REGEXP,
+			},
+			err: "",
 		}, {
 			m: &pb.Matcher{
 				Name:    "00",
@@ -824,6 +1161,13 @@ func TestValidateMatcher(t *testing.T) {
 				Type:    pb.Matcher_REGEXP,
 			},
 			err: "invalid regular expression",
+		}, {
+			m: &pb.Matcher{
+				Name:    "a",
+				Pattern: "))",
+				Type:    pb.Matcher_NOT_REGEXP,
+			},
+			err: "invalid regular expression",
 		}, {
 			m: &pb.Matcher{
 				Name:    "a",
@@ -842,26 +1186,13 @@ func TestValidateMatcher(t *testing.T) {
 	}
 
 	for _, c := range cases {
-		err := validateMatcher(c.m)
-		if err == nil {
-			if c.err != "" {
-				t.Errorf("expected error containing %q but got none", c.err)
-			}
-			continue
-		}
-		if err != nil && c.err == "" {
-			t.Errorf("unexpected error %q", err)
-			continue
-		}
-		if !strings.Contains(err.Error(), c.err) {
-			t.Errorf("expected error to contain %q but got %q", c.err, err)
-		}
+		checkErr(t, c.err, ValidateMatcher(c.m))
 	}
 }
 
 func TestValidateSilence(t *testing.T) {
 	var (
-		now            = utcNow()
+		now            = time.Now().UTC()
 		zeroTimestamp  = time.Time{}
 		validTimestamp = now
 	)
@@ -873,7 +1204,7 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "some_id",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
+					{Name: "a", Pattern: "b"},
 				},
 				StartsAt:  validTimestamp,
 				EndsAt:    validTimestamp,
@@ -885,7 +1216,7 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
+					{Name: "a", Pattern: "b"},
 				},
 				StartsAt:  validTimestamp,
 				EndsAt:    validTimestamp,
@@ -907,8 +1238,8 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "some_id",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
-					&pb.Matcher{Name: "00", Pattern: "b"},
+					{Name: "a", Pattern: "b"},
+					{Name: "00", Pattern: "b"},
 				},
 				StartsAt:  validTimestamp,
 				EndsAt:    validTimestamp,
@@ -920,7 +1251,20 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "some_id",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
+					{Name: "a", Pattern: ""},
+					{Name: "b", Pattern: ".*", Type: pb.Matcher_REGEXP},
+				},
+				StartsAt:  validTimestamp,
+				EndsAt:    validTimestamp,
+				UpdatedAt: validTimestamp,
+			},
+			err: "at least one matcher must not match the empty string",
+		},
+		{
+			s: &pb.Silence{
+				Id: "some_id",
+				Matchers: []*pb.Matcher{
+					{Name: "a", Pattern: "b"},
 				},
 				StartsAt:  now,
 				EndsAt:    now.Add(-time.Second),
@@ -932,7 +1276,7 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "some_id",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
+					{Name: "a", Pattern: "b"},
 				},
 				StartsAt:  zeroTimestamp,
 				EndsAt:    validTimestamp,
@@ -944,7 +1288,7 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "some_id",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
+					{Name: "a", Pattern: "b"},
 				},
 				StartsAt:  validTimestamp,
 				EndsAt:    zeroTimestamp,
@@ -956,7 +1300,7 @@ func TestValidateSilence(t *testing.T) {
 			s: &pb.Silence{
 				Id: "some_id",
 				Matchers: []*pb.Matcher{
-					&pb.Matcher{Name: "a", Pattern: "b"},
+					{Name: "a", Pattern: "b"},
 				},
 				StartsAt:  validTimestamp,
 				EndsAt:    validTimestamp,
@@ -966,61 +1310,52 @@ func TestValidateSilence(t *testing.T) {
 		},
 	}
 	for _, c := range cases {
-		err := validateSilence(c.s)
-		if err == nil {
-			if c.err != "" {
-				t.Errorf("expected error containing %q but got none", c.err)
-			}
-			continue
-		}
-		if err != nil && c.err == "" {
-			t.Errorf("unexpected error %q", err)
-			continue
-		}
-		if !strings.Contains(err.Error(), c.err) {
-			t.Errorf("expected error to contain %q but got %q", c.err, err)
-		}
+		checkErr(t, c.err, validateSilence(c.s))
 	}
 }
 
 func TestStateMerge(t *testing.T) {
-	now := utcNow()
+	now := time.Now().UTC()
 
 	// We only care about key names and timestamps for the
 	// merging logic.
-	newSilence := func(id string, ts time.Time) *pb.MeshSilence {
+	newSilence := func(id string, ts, exp time.Time) *pb.MeshSilence {
 		return &pb.MeshSilence{
-			Silence: &pb.Silence{Id: id, UpdatedAt: ts},
+			Silence:   &pb.Silence{Id: id, UpdatedAt: ts},
+			ExpiresAt: exp,
 		}
 	}
 
+	exp := now.Add(time.Minute)
+
 	cases := []struct {
 		a, b  state
 		final state
 	}{
 		{
 			a: state{
-				"a1": newSilence("a1", now),
-				"a2": newSilence("a2", now),
-				"a3": newSilence("a3", now),
+				"a1": newSilence("a1", now, exp),
+				"a2": newSilence("a2", now, exp),
+				"a3": newSilence("a3", now, exp),
 			},
 			b: state{
-				"b1": newSilence("b1", now),                   // new key, should be added
-				"a2": newSilence("a2", now.Add(-time.Minute)), // older timestamp, should be dropped
-				"a3": newSilence("a3", now.Add(time.Minute)),  // newer timestamp, should overwrite
+				"b1": newSilence("b1", now, exp),                                          // new key, should be added
+				"a2": newSilence("a2", now.Add(-time.Minute), exp),                        // older timestamp, should be dropped
+				"a3": newSilence("a3", now.Add(time.Minute), exp),                         // newer timestamp, should overwrite
+				"a4": newSilence("a4", now.Add(-time.Minute), now.Add(-time.Millisecond)), // new key, expired, should not be added
 			},
 			final: state{
-				"a1": newSilence("a1", now),
-				"a2": newSilence("a2", now),
-				"a3": newSilence("a3", now.Add(time.Minute)),
-				"b1": newSilence("b1", now),
+				"a1": newSilence("a1", now, exp),
+				"a2": newSilence("a2", now, exp),
+				"a3": newSilence("a3", now.Add(time.Minute), exp),
+				"b1": newSilence("b1", now, exp),
 			},
 		},
 	}
 
 	for _, c := range cases {
 		for _, e := range c.b {
-			c.a.merge(e)
+			c.a.merge(e, now)
 		}
 
 		require.Equal(t, c.final, c.a, "Merge result should match expectation")
@@ -1029,7 +1364,7 @@ func TestStateMerge(t *testing.T) {
 
 func TestStateCoding(t *testing.T) {
 	// Check whether encoding and decoding the data is symmetric.
-	now := utcNow()
+	now := time.Now().UTC()
 
 	cases := []struct {
 		entries []*pb.MeshSilence
@@ -1061,6 +1396,19 @@ func TestStateCoding(t *testing.T) {
 					},
 					ExpiresAt: now.Add(24 * time.Hour),
 				},
+				{
+					Silence: &pb.Silence{
+						Id: "3dfb2528-59ce-41eb-b465-f875a4e744a4",
+						Matchers: []*pb.Matcher{
+							{Name: "label1", Pattern: "val1", Type: pb.Matcher_NOT_EQUAL},
+							{Name: "label2", Pattern: "val.+", Type: pb.Matcher_NOT_REGEXP},
+						},
+						StartsAt:  now,
+						EndsAt:    now,
+						UpdatedAt: now,
+					},
+					ExpiresAt: now,
+				},
 			},
 		},
 	}
@@ -1091,3 +1439,73 @@ func TestStateDecodingError(t *testing.T) {
 	_, err = decodeState(bytes.NewReader(msg))
 	require.Equal(t, ErrInvalidState, err)
 }
+
+func benchmarkSilencesQuery(b *testing.B, numSilences int) {
+	s, err := New(Options{})
+	require.NoError(b, err)
+
+	clock := clock.NewMock()
+	s.clock = clock
+	now := clock.Now()
+
+	lset := model.LabelSet{"aaaa": "AAAA", "bbbb": "BBBB", "cccc": "CCCC"}
+
+	s.st = state{}
+	for i := 0; i < numSilences; i++ {
+		id := fmt.Sprint("ID", i)
+		// Patterns also contain the ID to bust any caches that might be used under the hood.
+		patA := "A{4}|" + id
+		patB := id // Does not match.
+		if i%10 == 0 {
+			// Every 10th time, have an actually matching pattern.
+			patB = "B(B|C)B.|" + id
+		}
+
+		s.st[id] = &pb.MeshSilence{Silence: &pb.Silence{
+			Id: id,
+			Matchers: []*pb.Matcher{
+				{Type: pb.Matcher_REGEXP, Name: "aaaa", Pattern: patA},
+				{Type: pb.Matcher_REGEXP, Name: "bbbb", Pattern: patB},
+			},
+			StartsAt:  now.Add(-time.Minute),
+			EndsAt:    now.Add(time.Hour),
+			UpdatedAt: now.Add(-time.Hour),
+		}}
+	}
+
+	// Run things once to populate the matcherCache.
+	sils, _, err := s.Query(
+		QState(types.SilenceStateActive),
+		QMatches(lset),
+	)
+	require.NoError(b, err)
+	require.Equal(b, numSilences/10, len(sils))
+
+	b.ResetTimer()
+	for i := 0; i < b.N; i++ {
+		sils, _, err := s.Query(
+			QState(types.SilenceStateActive),
+			QMatches(lset),
+		)
+		require.NoError(b, err)
+		require.Equal(b, numSilences/10, len(sils))
+	}
+}
+
+func Benchmark100SilencesQuery(b *testing.B) {
+	benchmarkSilencesQuery(b, 100)
+}
+
+func Benchmark1000SilencesQuery(b *testing.B) {
+	benchmarkSilencesQuery(b, 1000)
+}
+
+func Benchmark10000SilencesQuery(b *testing.B) {
+	benchmarkSilencesQuery(b, 10000)
+}
+
+// runtime.Gosched() does not "suspend" the current goroutine so there's no guarantee that the main goroutine won't
+// be able to continue. For more see https://pkg.go.dev/runtime#Gosched.
+func gosched() {
+	time.Sleep(1 * time.Millisecond)
+}
